from dataclasses import dataclass
from typing import Optional

import pytest

from mashumaro.jsonschema.builder import JSONSchemaBuilder, build_json_schema
from mashumaro.jsonschema.models import (
    Context,
    JSONSchema,
    JSONSchemaInstanceType,
)
from mashumaro.jsonschema.plugins import BasePlugin, DocstringDescriptionPlugin
from mashumaro.jsonschema.schema import Instance


class ThirdPartyType:
    pass


@dataclass
class DataClassWithDocstring:
    """Here is the docstring"""

    x: int


@dataclass
class DataClassWithoutDocstring:
    x: int


def test_basic_plugin():
    assert build_json_schema(int, plugins=[BasePlugin()]) == JSONSchema(
        type=JSONSchemaInstanceType.INTEGER
    )


def test_plugin_with_not_implemented_error():
    class NotImplementedErrorPlugin(BasePlugin):
        def get_schema(
            self,
            instance: Instance,
            ctx: Context,
            schema: Optional[JSONSchema] = None,
        ) -> Optional[JSONSchema]:
            raise NotImplementedError

    assert build_json_schema(
        int, plugins=[NotImplementedErrorPlugin()]
    ) == JSONSchema(type=JSONSchemaInstanceType.INTEGER)
    assert JSONSchemaBuilder(plugins=[NotImplementedErrorPlugin()]).build(
        int
    ) == JSONSchema(type=JSONSchemaInstanceType.INTEGER)


@pytest.mark.parametrize(
    ("obj", "docstring"),
    (
        (DataClassWithDocstring, "Here is the docstring"),
        (DataClassWithoutDocstring, "DataClassWithoutDocstring(x: int)"),
        (int, None),
    ),
)
def test_docstring_description_plugin(obj, docstring):
    assert build_json_schema(obj).description is None
    assert JSONSchemaBuilder().build(obj).description is None

    assert (
        build_json_schema(
            obj, plugins=[DocstringDescriptionPlugin()]
        ).description
        == docstring
    )
    assert (
        JSONSchemaBuilder(plugins=[DocstringDescriptionPlugin()])
        .build(obj)
        .description
        == docstring
    )


def test_third_party_type_plugin():
    third_party_json_schema = JSONSchema()

    class ThirdPartyTypePlugin(BasePlugin):
        def get_schema(
            self,
            instance: Instance,
            ctx: Context,
            schema: Optional[JSONSchema] = None,
        ) -> Optional[JSONSchema]:
            try:
                if issubclass(instance.type, ThirdPartyType):
                    return third_party_json_schema
            except TypeError:
                pass

    assert (
        build_json_schema(ThirdPartyType, plugins=[ThirdPartyTypePlugin()])
        is third_party_json_schema
    )
    assert (
        JSONSchemaBuilder(plugins=[ThirdPartyTypePlugin()]).build(
            ThirdPartyType
        )
        is third_party_json_schema
    )
    assert (
        JSONSchemaBuilder(plugins=[ThirdPartyTypePlugin()])
        .build(list[ThirdPartyType])
        .items
        is third_party_json_schema
    )
    assert (
        build_json_schema(
            list[ThirdPartyType], plugins=[ThirdPartyTypePlugin()]
        ).items
        is third_party_json_schema
    )
    with pytest.raises(NotImplementedError):
        build_json_schema(ThirdPartyType)
    with pytest.raises(NotImplementedError):
        JSONSchemaBuilder().build(ThirdPartyType)
    with pytest.raises(NotImplementedError):
        build_json_schema(list[ThirdPartyType])
    with pytest.raises(NotImplementedError):
        JSONSchemaBuilder().build(list[ThirdPartyType])



from dataclasses import dataclass

import pytest
import typing_extensions

from mashumaro import DataClassDictMixin
from mashumaro.core.meta.helpers import is_not_required, is_required
from mashumaro.exceptions import InvalidFieldValue


class TypedDictCorrectNotRequired(typing_extensions.TypedDict):
    required: int
    not_required: typing_extensions.NotRequired[int]


class TypedDictCorrectRequired(typing_extensions.TypedDict, total=False):
    required: typing_extensions.Required[int]
    not_required: int


def test_is_required():
    assert is_required(typing_extensions.Required[int])
    assert not is_required(typing_extensions.Self)


def test_is_not_required():
    assert is_not_required(typing_extensions.NotRequired[int])
    assert not is_not_required(typing_extensions.Self)


def test_typed_dict_correct_not_required():
    @dataclass
    class MyClass(DataClassDictMixin):
        x: TypedDictCorrectNotRequired

    obj = MyClass({"required": 42})
    assert MyClass.from_dict({"x": {"required": 42}}) == obj
    assert obj.to_dict() == {"x": {"required": 42}}


def test_typed_dict_correct_required():
    @dataclass
    class MyClass(DataClassDictMixin):
        x: TypedDictCorrectRequired

    with pytest.raises(InvalidFieldValue) as exc_info:
        MyClass.from_dict({"x": {}})
    assert exc_info.type is InvalidFieldValue
    assert exc_info.value.holder_class is MyClass
    assert exc_info.value.field_type is TypedDictCorrectRequired
    assert exc_info.value.field_value == {}
    assert exc_info.value.field_name == "x"



import json
from datetime import date
from typing import List

from mashumaro.codecs.json import (
    JSONDecoder,
    JSONEncoder,
    json_decode,
    json_encode,
)
from mashumaro.dialect import Dialect


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
    }


def test_json_decode():
    assert json_decode('["2023-09-22", "2023-09-23"]', List[date]) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_json_encode():
    assert (
        json_encode([date(2023, 9, 22), date(2023, 9, 23)], List[date])
        == '["2023-09-22", "2023-09-23"]'
    )


def test_decoder_with_default_dialect():
    decoder = JSONDecoder(List[date], default_dialect=MyDialect)
    assert decoder.decode("[738785, 738786]") == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_encoder_with_default_dialect():
    encoder = JSONEncoder(List[date], default_dialect=MyDialect)
    assert (
        encoder.encode([date(2023, 9, 22), date(2023, 9, 23)])
        == "[738785, 738786]"
    )


def test_pre_decoder_func():
    decoder = JSONDecoder(
        List[date],
        pre_decoder_func=lambda v: json.loads(
            "[" + ",".join(f'"{v}"' for v in v.split(",")) + "]"
        ),
    )
    assert decoder.decode("2023-09-22,2023-09-23") == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_post_encoder_func():
    encoder = JSONEncoder(
        List[date],
        post_encoder_func=lambda v: json.dumps(v, separators=(",", ":")),
    )
    assert (
        encoder.encode(
            [
                date(2023, 9, 22),
                date(2023, 9, 23),
            ]
        )
        == '["2023-09-22","2023-09-23"]'
    )



import json
from dataclasses import dataclass
from datetime import datetime
from typing import List

import pytest

from mashumaro.exceptions import MissingField
from mashumaro.mixins.json import DataClassJSONMixin

from .entities import MyEnum


def test_to_json():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: List[int]

    dumped = json.dumps({"x": [1, 2, 3]})
    assert DataClass([1, 2, 3]).to_json() == dumped


def test_from_json():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: List[int]

    dumped = json.dumps({"x": [1, 2, 3]})
    assert DataClass.from_json(dumped) == DataClass([1, 2, 3])


def test_to_json_with_custom_encoder():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: List[int]
        x_count: int

    def encoder(d):
        dd = dict(d)
        dd.pop("x_count", None)
        return json.dumps(dd)

    instance = DataClass(x=[1, 2, 3], x_count=3)
    dumped = json.dumps({"x": [1, 2, 3]})
    assert instance.to_json(encoder=encoder) == dumped
    assert instance.to_json() != dumped


def test_from_json_with_custom_decoder():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: List[int]
        x_count: int

    def decoder(s):
        d = json.loads(s)
        d["x_count"] = len(d.get("x", []))
        return d

    instance = DataClass(x=[1, 2, 3], x_count=3)
    dumped = json.dumps({"x": [1, 2, 3]})
    assert DataClass.from_json(dumped, decoder=decoder) == instance
    with pytest.raises(MissingField):
        assert DataClass.from_json(dumped)


def test_json_enum():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: MyEnum

    dumped = '{"x": "letter a"}'
    instance = DataClass(MyEnum.a)

    assert instance.to_json() == dumped
    assert instance.to_json() == dumped
    assert DataClass.from_json(dumped) == instance


def test_json_bytes():
    @dataclass
    class DataClass(DataClassJSONMixin):
        x: bytes

    dumped = r'{"x": "MTIz\n"}'
    instance = DataClass(b"123")

    assert instance.to_json() == dumped
    assert DataClass.from_json(dumped) == instance


def test_json_datetime():
    dt = datetime(2018, 10, 29, 12, 46, 55, 308495)

    @dataclass
    class DataClass(DataClassJSONMixin):
        x: datetime

    dumped = json.dumps({"x": dt.isoformat()})
    instance = DataClass(x=dt)

    assert instance.to_json() == dumped
    assert DataClass.from_json(dumped) == instance



import re
import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, field, replace
from functools import cached_property
from types import new_class
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar, Union

from typing_extensions import ParamSpec, TypeAlias

from mashumaro.core.meta.code.lines import CodeLines
from mashumaro.core.meta.helpers import (
    get_type_origin,
    is_annotated,
    is_generic,
    is_hashable_type,
    is_self,
    type_name,
)
from mashumaro.exceptions import UnserializableField

if TYPE_CHECKING:  # pragma: no cover
    from mashumaro.core.meta.code.builder import CodeBuilder
else:
    CodeBuilder = Any


class TypeMatchEligibleExpression(str):
    pass


NoneType = type(None)
Expression: TypeAlias = Union[str, TypeMatchEligibleExpression]

P = ParamSpec("P")
T = TypeVar("T")

_PY_VALID_ID_RE = re.compile(r"\W|^(?=\d)")


class AttrsHolder:
    def __new__(
        cls, name: Optional[str] = None, *args: Any, **kwargs: Any
    ) -> Any:
        ah = new_class("AttrsHolder")
        ah_id = id(ah)
        if not name:
            name = f"attrs_{ah_id}"
        ah.__name__ = ah.__qualname__ = name
        return ah


class ExpressionWrapper:
    def __init__(self, expression: str):
        self.expression = expression


@dataclass
class FieldContext:
    name: str
    metadata: Mapping
    packer: Optional[str] = None
    unpacker: Optional[str] = None

    def copy(self, **changes: Any) -> "FieldContext":
        return replace(self, **changes)


@dataclass
class ValueSpec:
    type: Type
    origin_type: Type = field(init=False)
    expression: Expression
    builder: CodeBuilder
    field_ctx: FieldContext
    could_be_none: bool = True
    annotated_type: Optional[Type] = None
    owner: Optional[Type] = None
    no_copy_collections: Sequence = tuple()

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "type":
            self.origin_type = get_type_origin(value)
        super().__setattr__(key, value)

    def copy(self, **changes: Any) -> "ValueSpec":
        return replace(self, **changes)

    @cached_property
    def annotations(self) -> Sequence[str]:
        return getattr(self.annotated_type, "__metadata__", [])

    @cached_property
    def attrs(self) -> Any:
        if self.builder.is_nailed:
            return self.builder.attrs
        if is_self(self.type):
            typ = self.builder.cls
        else:
            typ = self.origin_type
        attrs = self.attrs_registry.get(typ)
        if attrs is None:
            attrs = AttrsHolder()
            self.attrs_registry[typ] = attrs
        return attrs

    @cached_property
    def cls_attrs_name(self) -> str:
        if self.builder.is_nailed:
            return "cls"
        else:
            self.builder.ensure_object_imported(self.attrs)
            return self.attrs.__name__

    @cached_property
    def self_attrs_name(self) -> str:
        if self.builder.is_nailed:
            return "self"
        else:
            self.builder.ensure_object_imported(self.attrs)
            return self.attrs.__name__

    @cached_property
    def attrs_registry(self) -> dict[Any, Any]:
        return self.builder.attrs_registry

    @cached_property
    def attrs_registry_name(self) -> str:
        name = f"attrs_registry_{id(self.attrs_registry)}"
        self.builder.ensure_object_imported(self.attrs_registry, name)
        return name


class AbstractMethodBuilder(ABC):
    @abstractmethod
    def get_method_prefix(self) -> str:  # pragma: no cover
        raise NotImplementedError

    def _generate_method_name(
        self, spec: ValueSpec
    ) -> str:  # pragma: no cover
        prefix = self.get_method_prefix()
        if prefix:
            prefix = f"{prefix}_"
        if spec.field_ctx.name:
            suffix = f"_{spec.field_ctx.name}"
        else:
            suffix = ""
        return f"__{prefix}{spec.builder.cls.__name__}{suffix}__{random_hex()}"

    @abstractmethod
    def _add_definition(self, spec: ValueSpec, lines: CodeLines) -> str:
        raise NotImplementedError

    @abstractmethod
    def _generate_method_args(self, spec: ValueSpec) -> str:
        raise NotImplementedError

    @abstractmethod
    def _add_body(
        self, spec: ValueSpec, lines: CodeLines
    ) -> None:  # pragma: no cover
        raise NotImplementedError

    def _add_setattr(
        self, spec: ValueSpec, method_name: str, lines: CodeLines
    ) -> None:
        lines.append(
            f"setattr({spec.cls_attrs_name}, '{method_name}', {method_name})"
        )

    def _compile(self, spec: ValueSpec, lines: CodeLines) -> None:
        if spec.builder.get_config().debug:
            print(f"{type_name(spec.builder.cls)}:")
            print(lines.as_text())
        exec(lines.as_text(), spec.builder.globals, spec.builder.__dict__)

    @abstractmethod
    def _get_call_expr(self, spec: ValueSpec, method_name: str) -> str:
        raise NotImplementedError

    def _before_build(self, spec: ValueSpec) -> None:
        pass

    def _get_existing_method(self, spec: ValueSpec) -> Optional[str]:
        return None

    def build(self, spec: ValueSpec) -> str:
        self._before_build(spec)
        if method := self._get_existing_method(spec):
            return method
        lines = CodeLines()
        method_name = self._add_definition(spec, lines)
        with lines.indent():
            self._add_body(spec, lines)
        self._add_setattr(spec, method_name, lines)
        self._compile(spec, lines)
        return self._get_call_expr(spec, method_name)


ValueSpecExprCreator: TypeAlias = Callable[[ValueSpec], Optional[Expression]]


@dataclass
class Registry:
    _registry: list[ValueSpecExprCreator] = field(default_factory=list)

    def register(self, function: ValueSpecExprCreator) -> ValueSpecExprCreator:
        self._registry.append(function)
        return function

    def get(self, spec: ValueSpec) -> Expression:
        if is_annotated(spec.type):
            spec.annotated_type = spec.builder.get_real_type(
                spec.field_ctx.name, spec.type
            )
            spec.type = get_type_origin(spec.type)
        spec.type = spec.builder.get_real_type(spec.field_ctx.name, spec.type)
        spec.builder.add_type_modules(spec.type)
        for packer in self._registry:
            expr = packer(spec)
            if expr is not None:
                return expr
        raise UnserializableField(
            spec.field_ctx.name, spec.type, spec.builder.cls
        )


def ensure_generic_collection(spec: ValueSpec) -> bool:
    if not is_generic(spec.type):
        return False
    return True


def ensure_mapping_key_type_hashable(
    spec: ValueSpec, type_args: Sequence[Type]
) -> bool:
    if type_args:
        first_type_arg = type_args[0]
        if not is_hashable_type(first_type_arg):
            raise UnserializableField(
                field_name=spec.field_ctx.name,
                field_type=spec.type,
                holder_class=spec.builder.cls,
                msg=(
                    f"{type_name(first_type_arg, short=True)} "
                    "is unhashable and can not be used as a key"
                ),
            )
    return True


def ensure_generic_collection_subclass(
    spec: ValueSpec, *checked_types: Type
) -> bool:
    return issubclass(
        spec.origin_type, checked_types
    ) and ensure_generic_collection(spec)


def ensure_generic_mapping(
    spec: ValueSpec, args: Sequence[Type], checked_type: Type
) -> bool:
    return ensure_generic_collection_subclass(
        spec, checked_type
    ) and ensure_mapping_key_type_hashable(spec, args)


def expr_or_maybe_none(spec: ValueSpec, new_expr: Expression) -> Expression:
    if spec.could_be_none:
        return f"{new_expr} if {spec.expression} is not None else None"
    else:
        return new_expr


def random_hex() -> str:
    return str(uuid.uuid4().hex)


def clean_id(value: str) -> str:
    if not value:
        return "_"

    return _PY_VALID_ID_RE.sub("_", value)



from dataclasses import dataclass

import pytest
from typing_extensions import Literal

from mashumaro import DataClassDictMixin
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.exceptions import InvalidFieldValue
from mashumaro.helper import pass_through
from tests.entities import MyEnum


def test_literal_with_str():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal["1", "2", "3"]

    assert DataClass.from_dict({"x": "1"}) == DataClass("1")
    assert DataClass.from_dict({"x": "2"}) == DataClass("2")

    assert DataClass("1").to_dict() == {"x": "1"}
    assert DataClass("2").to_dict() == {"x": "2"}

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": 1})
    with pytest.raises(InvalidFieldValue):
        DataClass(1).to_dict()


def test_literal_with_int():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[1, 2]

    assert DataClass.from_dict({"x": 1}) == DataClass(1)
    assert DataClass.from_dict({"x": 2}) == DataClass(2)

    assert DataClass(1).to_dict() == {"x": 1}
    assert DataClass(2).to_dict() == {"x": 2}

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "1"})
    with pytest.raises(InvalidFieldValue):
        DataClass("1").to_dict()


def test_literal_with_bool():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[True, False]

    assert DataClass.from_dict({"x": True}) == DataClass(True)
    assert DataClass.from_dict({"x": False}) == DataClass(False)

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "a"})
    with pytest.raises(InvalidFieldValue):
        DataClass("a").to_dict()


def test_literal_with_none():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[None]

    assert DataClass.from_dict({"x": None}) == DataClass(None)
    assert DataClass(None).to_dict() == {"x": None}
    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "1"})
    with pytest.raises(InvalidFieldValue):
        DataClass("1").to_dict()


def test_literal_with_bytes():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[b"\x00"]

    assert DataClass.from_dict({"x": "AA==\n"}) == DataClass(b"\x00")
    assert DataClass(b"\x00").to_dict() == {"x": "AA==\n"}
    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "\x00"})
    with pytest.raises(InvalidFieldValue):
        DataClass("AA==\n").to_dict()


def test_literal_with_bytes_overridden():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[b"\x00"]

        class Config(BaseConfig):
            serialization_strategy = {bytes: pass_through}

    assert DataClass.from_dict({"x": b"\x00"}) == DataClass(b"\x00")
    assert DataClass(b"\x00").to_dict() == {"x": b"\x00"}
    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "AA==\n"})


def test_literal_with_enum():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[MyEnum.a]

    assert DataClass.from_dict({"x": "letter a"}) == DataClass(MyEnum.a)
    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "letter b"})
    with pytest.raises(InvalidFieldValue):
        DataClass(MyEnum.b).to_dict()


def test_literal_with_dialect():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Literal[b"\x00"]

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]

    class MyDialect(Dialect):
        serialization_strategy = {bytes: pass_through}

    instance = DataClass(b"\x00")
    assert DataClass.from_dict({"x": b"\x00"}, dialect=MyDialect) == instance
    assert instance.to_dict(dialect=MyDialect) == {"x": b"\x00"}
    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "AA==\n"}, dialect=MyDialect)



from dataclasses import dataclass
from datetime import date
from typing import Optional

import pytest
from typing_extensions import Literal

from mashumaro import DataClassDictMixin
from mashumaro.codecs.basic import decode
from mashumaro.config import BaseConfig
from mashumaro.exceptions import (
    InvalidFieldValue,
    SuitableVariantNotFoundError,
)
from mashumaro.types import Discriminator

DT_STR = "2023-05-30"
DT_DATE = date(2023, 5, 30)

X_1 = {"x": DT_STR, "type": 1}
X_2 = {"x": DT_STR, "type": 2}
X_3 = {"x": DT_STR, "type": 3}


@dataclass
class VariantBySubtypes(DataClassDictMixin):
    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class _VariantBySubtypes:
    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class VariantBySubtypesSub1(VariantBySubtypes):
    x: str
    type: Literal[1] = 1


@dataclass
class _VariantBySubtypesSub1(_VariantBySubtypes):
    x: str
    type: Literal[1] = 1


@dataclass
class VariantBySubtypesSub2(VariantBySubtypesSub1):
    x: date
    type: Literal[2] = 2


@dataclass
class _VariantBySubtypesSub2(_VariantBySubtypesSub1):
    x: date
    type: Literal[2] = 2


@dataclass
class VariantBySubtypesSub3(VariantBySubtypes):
    x: date
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class _VariantBySubtypesSub3(_VariantBySubtypes):
    x: date
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class VariantBySubtypesSub4(VariantBySubtypesSub3):
    pass


@dataclass
class _VariantBySubtypesSub4(_VariantBySubtypesSub3):
    pass


@dataclass
class VariantByFieldWithSubtypes(DataClassDictMixin):
    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class _VariantByFieldWithSubtypes:
    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class VariantByFieldWithSubtypesSub1(VariantByFieldWithSubtypes):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class _VariantByFieldWithSubtypesSub1(_VariantByFieldWithSubtypes):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class VariantByFieldWithSubtypesSub2(VariantByFieldWithSubtypesSub1):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class _VariantByFieldWithSubtypesSub2(_VariantByFieldWithSubtypesSub1):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class VariantByFieldWithSubtypesSub3(VariantByFieldWithSubtypes):
    x: Optional[date] = None
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class _VariantByFieldWithSubtypesSub3(_VariantByFieldWithSubtypes):
    x: Optional[date] = None
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class VariantByFieldWithSubtypesSub4(VariantByFieldWithSubtypesSub3):
    pass


@dataclass
class _VariantByFieldWithSubtypesSub4(_VariantByFieldWithSubtypesSub3):
    pass


@dataclass
class VariantByFieldWithSupertypesAndSubtypes(DataClassDictMixin):
    type: Literal["unknown"] = "unknown"

    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        )


@dataclass
class _VariantByFieldWithSupertypesAndSubtypes:
    type: Literal["unknown"] = "unknown"

    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        )


@dataclass
class VariantByFieldWithSupertypesAndSubtypesSub1(
    VariantByFieldWithSupertypesAndSubtypes
):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class _VariantByFieldWithSupertypesAndSubtypesSub1(
    _VariantByFieldWithSupertypesAndSubtypes
):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class VariantByFieldWithSupertypesAndSubtypesSub2(
    VariantByFieldWithSupertypesAndSubtypesSub1
):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class _VariantByFieldWithSupertypesAndSubtypesSub2(
    _VariantByFieldWithSupertypesAndSubtypesSub1
):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class VariantByFieldWithSupertypesAndSubtypesSub3(
    VariantByFieldWithSupertypesAndSubtypes
):
    x: Optional[date] = None
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        )


@dataclass
class _VariantByFieldWithSupertypesAndSubtypesSub3(
    _VariantByFieldWithSupertypesAndSubtypes
):
    x: Optional[date] = None
    type: Literal[3] = 3

    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        )


@dataclass
class VariantByFieldWithSupertypesAndSubtypesSub4(
    VariantByFieldWithSupertypesAndSubtypesSub3
):
    pass


@dataclass
class _VariantByFieldWithSupertypesAndSubtypesSub4(
    _VariantByFieldWithSupertypesAndSubtypesSub3
):
    pass


@dataclass
class VariantBySupertypesAndSubtypes(DataClassDictMixin):
    type: Literal["unknown"] = "unknown"

    class Config(BaseConfig):
        discriminator = Discriminator(
            include_supertypes=True, include_subtypes=True
        )


@dataclass
class _VariantBySupertypesAndSubtypes:
    type: Literal["unknown"] = "unknown"

    class Config(BaseConfig):
        discriminator = Discriminator(
            include_supertypes=True, include_subtypes=True
        )


@dataclass
class VariantBySupertypesAndSubtypesSub1(VariantBySupertypesAndSubtypes):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class _VariantBySupertypesAndSubtypesSub1(_VariantBySupertypesAndSubtypes):
    x: Optional[str] = None
    type: Literal[1] = 1


@dataclass
class VariantBySupertypesAndSubtypesSub2(VariantBySupertypesAndSubtypesSub1):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class _VariantBySupertypesAndSubtypesSub2(_VariantBySupertypesAndSubtypesSub1):
    x: Optional[date] = None
    type: Literal[2] = 2


@dataclass
class Foo1(DataClassDictMixin):
    x1: int

    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class _Foo1:
    x1: int

    class Config(BaseConfig):
        discriminator = Discriminator(field="type", include_subtypes=True)


@dataclass
class Foo2(Foo1):
    x2: int


@dataclass
class _Foo2(_Foo1):
    x2: int


@dataclass
class Foo3(Foo2):
    x: int
    type = 3


@dataclass
class _Foo3(_Foo2):
    x: int
    type = 3


@dataclass
class Foo4(Foo2):
    x: int
    type = 4


@dataclass
class _Foo4(_Foo2):
    x: int
    type = 4


@dataclass
class Bar1(DataClassDictMixin):
    x1: int

    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class _Bar1:
    x1: int

    class Config(BaseConfig):
        discriminator = Discriminator(include_subtypes=True)


@dataclass
class Bar2(Bar1):
    x2: int


@dataclass
class _Bar2(_Bar1):
    x2: int


@dataclass
class Bar3(Bar2):
    x: int
    type = 3


@dataclass
class _Bar3(_Bar2):
    x: int
    type = 3


@dataclass
class Bar4(Bar2):
    x: int
    type = 4


@dataclass
class _Bar4(_Bar2):
    x: int
    type = 4


@dataclass
class VariantWitCustomTagger(DataClassDictMixin):
    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: cls.__name__.lower(),
        )


@dataclass
class _VariantWitCustomTagger:
    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: cls.__name__.lower(),
        )


@dataclass
class VariantWitCustomTaggerSub1(VariantWitCustomTagger):
    pass


@dataclass
class _VariantWitCustomTaggerSub1(_VariantWitCustomTagger):
    pass


@dataclass
class VariantWitCustomTaggerSub2(VariantWitCustomTagger):
    pass


@dataclass
class _VariantWitCustomTaggerSub2(_VariantWitCustomTagger):
    pass


@dataclass
class VariantWithMultipleTags(DataClassDictMixin):
    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: [
                cls.__name__.lower(),
                cls.__name__.upper(),
            ],
        )


@dataclass
class VariantWithMultipleTagsOne(VariantWithMultipleTags):
    pass


@dataclass
class VariantWithMultipleTagsTwo(VariantWithMultipleTags):
    pass


@dataclass
class _VariantWithMultipleTags:
    class Config(BaseConfig):
        discriminator = Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: [
                cls.__name__.lower(),
                cls.__name__.upper(),
            ],
        )


@dataclass
class _VariantWithMultipleTagsOne(_VariantWithMultipleTags):
    pass


@dataclass
class _VariantWithMultipleTagsTwo(_VariantWithMultipleTags):
    pass


def test_by_subtypes():
    assert VariantBySubtypes.from_dict(X_1) == VariantBySubtypesSub1(x=DT_STR)
    assert decode(X_1, _VariantBySubtypes) == _VariantBySubtypesSub1(x=DT_STR)

    assert VariantBySubtypes.from_dict(X_2) == VariantBySubtypesSub2(x=DT_DATE)
    assert decode(X_2, _VariantBySubtypes) == _VariantBySubtypesSub2(x=DT_DATE)

    assert VariantBySubtypes.from_dict(X_3) == VariantBySubtypesSub4(DT_DATE)
    assert decode(X_3, _VariantBySubtypes) == _VariantBySubtypesSub4(DT_DATE)

    assert VariantBySubtypesSub3.from_dict(X_3) == VariantBySubtypesSub4(
        DT_DATE
    )
    assert decode(X_3, _VariantBySubtypesSub3) == _VariantBySubtypesSub4(
        DT_DATE
    )

    with pytest.raises(SuitableVariantNotFoundError):
        VariantBySubtypesSub3.from_dict(X_1)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_1, _VariantBySubtypesSub3)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantBySubtypesSub3.from_dict(X_2)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_2, _VariantBySubtypesSub3)

    @dataclass
    class MyClass(DataClassDictMixin):
        x: VariantBySubtypes

    @dataclass
    class _MyClass:
        x: _VariantBySubtypes

    assert MyClass.from_dict({"x": X_1}) == MyClass(
        VariantBySubtypesSub1(DT_STR)
    )
    assert decode({"x": X_1}, _MyClass) == _MyClass(
        _VariantBySubtypesSub1(DT_STR)
    )

    assert MyClass.from_dict({"x": X_2}) == MyClass(
        VariantBySubtypesSub2(DT_DATE)
    )
    assert decode({"x": X_2}, _MyClass) == _MyClass(
        _VariantBySubtypesSub2(DT_DATE)
    )

    assert MyClass.from_dict({"x": X_3}) == MyClass(
        VariantBySubtypesSub4(DT_DATE)
    )
    assert decode({"x": X_3}, _MyClass) == _MyClass(
        _VariantBySubtypesSub4(DT_DATE)
    )

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": {}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {}}, _MyClass)


def test_by_supertypes():
    with pytest.raises(ValueError) as exc_info:

        @dataclass
        class VariantBySupertypes(DataClassDictMixin):
            class Config(BaseConfig):
                discriminator = Discriminator(include_supertypes=True)

    assert str(exc_info.value) == (
        "Config based discriminator must have 'include_subtypes' enabled"
    )


def test_by_field_with_subtypes():
    assert VariantByFieldWithSubtypes.from_dict(
        X_1
    ) == VariantByFieldWithSubtypesSub1(x=DT_STR)
    assert decode(
        X_1, _VariantByFieldWithSubtypes
    ) == _VariantByFieldWithSubtypesSub1(x=DT_STR)

    assert VariantByFieldWithSubtypes.from_dict(
        X_2
    ) == VariantByFieldWithSubtypesSub2(x=DT_DATE)
    assert decode(
        X_2, _VariantByFieldWithSubtypes
    ) == _VariantByFieldWithSubtypesSub2(x=DT_DATE)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSubtypes.from_dict(X_3)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_3, _VariantByFieldWithSubtypes)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSubtypesSub3.from_dict(X_3)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_3, _VariantByFieldWithSubtypesSub3)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSubtypesSub3.from_dict(X_1)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_1, _VariantByFieldWithSubtypesSub3)

    @dataclass
    class MyClass(DataClassDictMixin):
        x: VariantByFieldWithSubtypes

    @dataclass
    class _MyClass:
        x: _VariantByFieldWithSubtypes

    assert MyClass.from_dict({"x": X_1}) == MyClass(
        VariantByFieldWithSubtypesSub1(DT_STR)
    )
    assert decode({"x": X_1}, _MyClass) == _MyClass(
        _VariantByFieldWithSubtypesSub1(DT_STR)
    )

    assert MyClass.from_dict({"x": X_2}) == MyClass(
        VariantByFieldWithSubtypesSub2(DT_DATE)
    )
    assert decode({"x": X_2}, _MyClass) == _MyClass(
        _VariantByFieldWithSubtypesSub2(DT_DATE)
    )

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": X_3})
    with pytest.raises(InvalidFieldValue):
        decode({"x": X_3}, _MyClass)

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": {}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {}}, _MyClass)


def test_by_field_with_supertypes():
    with pytest.raises(ValueError) as exc_info:

        @dataclass
        class VariantByFieldWithSupertypes(DataClassDictMixin):
            class Config(BaseConfig):
                discriminator = Discriminator(
                    field="type", include_supertypes=True
                )

    assert str(exc_info.value) == (
        "Config based discriminator must have 'include_subtypes' enabled"
    )


def test_by_field_with_supertypes_and_subtypes():
    assert VariantByFieldWithSupertypesAndSubtypes.from_dict(
        X_1
    ) == VariantByFieldWithSupertypesAndSubtypesSub1(x=DT_STR)
    assert decode(
        X_1, _VariantByFieldWithSupertypesAndSubtypes
    ) == _VariantByFieldWithSupertypesAndSubtypesSub1(x=DT_STR)

    assert VariantByFieldWithSupertypesAndSubtypes.from_dict(
        X_2
    ) == VariantByFieldWithSupertypesAndSubtypesSub2(x=DT_DATE)
    assert decode(
        X_2, _VariantByFieldWithSupertypesAndSubtypes
    ) == _VariantByFieldWithSupertypesAndSubtypesSub2(x=DT_DATE)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSupertypesAndSubtypes.from_dict(X_3)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_3, _VariantByFieldWithSupertypesAndSubtypes)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSupertypesAndSubtypesSub3.from_dict(X_3)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_3, _VariantByFieldWithSupertypesAndSubtypesSub3)

    with pytest.raises(SuitableVariantNotFoundError):
        VariantByFieldWithSupertypesAndSubtypesSub3.from_dict(X_1)
    with pytest.raises(SuitableVariantNotFoundError):
        decode(X_1, _VariantByFieldWithSupertypesAndSubtypesSub3)

    @dataclass
    class MyClass(DataClassDictMixin):
        x: VariantByFieldWithSupertypesAndSubtypes

    @dataclass
    class _MyClass:
        x: _VariantByFieldWithSupertypesAndSubtypes

    assert MyClass.from_dict({"x": X_1}) == MyClass(
        VariantByFieldWithSupertypesAndSubtypesSub1(x=DT_STR)
    )
    assert decode({"x": X_1}, _MyClass) == _MyClass(
        _VariantByFieldWithSupertypesAndSubtypesSub1(x=DT_STR)
    )

    assert MyClass.from_dict({"x": X_2}) == MyClass(
        VariantByFieldWithSupertypesAndSubtypesSub2(x=DT_DATE)
    )
    assert decode({"x": X_2}, _MyClass) == _MyClass(
        _VariantByFieldWithSupertypesAndSubtypesSub2(x=DT_DATE)
    )

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": X_3})
    with pytest.raises(InvalidFieldValue):
        decode({"x": X_3}, _MyClass)

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": {"type": "unknown"}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {"type": "unknown"}}, _MyClass)

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": {}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {}}, _MyClass)

    with pytest.raises(InvalidFieldValue):
        MyClass.from_dict({"x": {"type": "invalid"}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {"type": "invalid"}}, _MyClass)


def test_by_supertypes_and_subtypes():
    assert VariantBySupertypesAndSubtypes.from_dict(
        X_1
    ) == VariantBySupertypesAndSubtypesSub1(x=DT_STR)
    assert decode(
        X_1, _VariantBySupertypesAndSubtypes
    ) == _VariantBySupertypesAndSubtypesSub1(x=DT_STR)

    assert VariantBySupertypesAndSubtypes.from_dict(
        X_2
    ) == VariantBySupertypesAndSubtypesSub2(x=DT_DATE)
    assert decode(
        X_2, _VariantBySupertypesAndSubtypes
    ) == _VariantBySupertypesAndSubtypesSub2(x=DT_DATE)

    @dataclass
    class MyClass(DataClassDictMixin):
        x: VariantBySupertypesAndSubtypes

    @dataclass
    class _MyClass:
        x: _VariantBySupertypesAndSubtypes

    assert MyClass.from_dict({"x": X_1}) == MyClass(
        VariantBySupertypesAndSubtypesSub1(x=DT_STR)
    )
    assert decode({"x": X_1}, _MyClass) == _MyClass(
        _VariantBySupertypesAndSubtypesSub1(x=DT_STR)
    )

    assert MyClass.from_dict({"x": X_2}) == MyClass(
        VariantBySupertypesAndSubtypesSub2(x=DT_DATE)
    )
    assert decode({"x": X_2}, _MyClass) == _MyClass(
        _VariantBySupertypesAndSubtypesSub2(x=DT_DATE)
    )

    assert MyClass.from_dict({"x": {}}) == MyClass(
        VariantBySupertypesAndSubtypesSub1()
    )
    assert decode({"x": {}}, _MyClass) == _MyClass(
        _VariantBySupertypesAndSubtypesSub1()
    )


def test_subclass_tree_with_class_without_field():
    assert Foo1.from_dict({"type": 3, "x1": 1, "x2": 2, "x": 42}) == Foo3(
        1, 2, 42
    )
    assert decode({"type": 3, "x1": 1, "x2": 2, "x": 42}, _Foo1) == _Foo3(
        1, 2, 42
    )

    assert Foo1.from_dict({"type": 4, "x1": 1, "x2": 2, "x": 42}) == Foo4(
        1, 2, 42
    )
    assert decode({"type": 4, "x1": 1, "x2": 2, "x": 42}, _Foo1) == _Foo4(
        1, 2, 42
    )

    assert Bar1.from_dict({"type": 3, "x1": 1, "x2": 2, "x": 42}) == Bar2(1, 2)
    assert decode({"type": 3, "x1": 1, "x2": 2, "x": 42}, _Bar1) == _Bar2(1, 2)
    assert Bar1.from_dict({"type": 4, "x1": 1, "x2": 2, "x": 42}) == Bar2(1, 2)
    assert decode({"type": 4, "x1": 1, "x2": 2, "x": 42}, _Bar1) == _Bar2(1, 2)


def test_by_subtypes_with_custom_variant_tagger():
    assert (
        VariantWitCustomTagger.from_dict(
            {"type": "variantwitcustomtaggersub1"}
        )
        == VariantWitCustomTaggerSub1()
    )
    assert (
        decode(
            {"type": "_variantwitcustomtaggersub1"}, _VariantWitCustomTagger
        )
        == _VariantWitCustomTaggerSub1()
    )

    assert (
        VariantWitCustomTagger.from_dict(
            {"type": "variantwitcustomtaggersub2"}
        )
        == VariantWitCustomTaggerSub2()
    )
    assert (
        decode(
            {"type": "_variantwitcustomtaggersub2"}, _VariantWitCustomTagger
        )
        == _VariantWitCustomTaggerSub2()
    )

    with pytest.raises(SuitableVariantNotFoundError):
        VariantWitCustomTagger.from_dict({"type": "unknown"})
    with pytest.raises(SuitableVariantNotFoundError):
        decode({"type": "unknown"}, _VariantWitCustomTagger)


def test_by_subtypes_with_custom_variant_tagger_and_multiple_tags():
    for variant in (VariantWithMultipleTagsOne, VariantWithMultipleTagsTwo):
        for tag in (variant.__name__.lower(), variant.__name__.upper()):
            assert (
                VariantWithMultipleTags.from_dict({"type": tag}) == variant()
            )
    for variant in (_VariantWithMultipleTagsOne, _VariantWithMultipleTagsTwo):
        for tag in (variant.__name__.lower(), variant.__name__.upper()):
            assert decode({"type": tag}, _VariantWithMultipleTags) == variant()

    with pytest.raises(SuitableVariantNotFoundError):
        VariantWithMultipleTags.from_dict({"type": "unknown"})
    with pytest.raises(SuitableVariantNotFoundError):
        decode({"type": "unknown"}, _VariantWithMultipleTags)



from dataclasses import dataclass, field

import pytest

from mashumaro import DataClassDictMixin, field_options
from mashumaro.core.const import PY_310_MIN

if not PY_310_MIN:
    pytest.skip("requires python>=3.10", allow_module_level=True)


def test_field_options_in_dataclass_with_slots():
    @dataclass(slots=True)
    class DataClass(DataClassDictMixin):
        x: int = field(metadata=field_options(serialize=str, alias="alias"))

    instance = DataClass(123)
    assert DataClass.from_dict({"alias": 123}) == instance
    assert instance.to_dict() == {"x": "123"}


def test_field_options_in_inherited_dataclass_with_slots():
    @dataclass
    class BaseDataClass(DataClassDictMixin):
        y: int

    @dataclass(slots=True)
    class DataClass(BaseDataClass):
        x: int = field(metadata=field_options(serialize=str, alias="alias"))

    instance = DataClass(x=123, y=456)
    assert DataClass.from_dict({"alias": 123, "y": 456}) == instance
    assert instance.to_dict() == {"x": "123", "y": 456}


def test_no_field_options_in_inherited_dataclass_with_slots():
    @dataclass
    class BaseDataClass(DataClassDictMixin):
        y: int

    @dataclass(slots=True)
    class DataClass(BaseDataClass):
        x: int

    instance = DataClass(x=123, y=456)
    assert DataClass.from_dict({"x": 123, "y": 456}) == instance
    assert instance.to_dict() == {"x": 123, "y": 456}


def test_no_field_options_in_inherited_dataclass_with_slots_and_default():
    @dataclass
    class BaseDataClass(DataClassDictMixin):
        y: int

    @dataclass(slots=True)
    class DataClass(BaseDataClass):
        x: int = 123

    instance = DataClass(y=456)
    assert DataClass.from_dict({"y": 456}) == instance
    assert instance.to_dict() == {"x": 123, "y": 456}



from datetime import date
from typing import List

import msgpack

from mashumaro.codecs.msgpack import (
    MessagePackDecoder,
    MessagePackEncoder,
    msgpack_decode,
    msgpack_encode,
)
from mashumaro.dialect import Dialect


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
    }


def test_msgpack_decode():
    data = msgpack.dumps(["2023-09-22", "2023-09-23"])
    assert msgpack_decode(data, List[date]) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_msgpack_encode():
    data = msgpack.dumps(["2023-09-22", "2023-09-23"])
    assert (
        msgpack_encode([date(2023, 9, 22), date(2023, 9, 23)], List[date])
        == data
    )


def test_decoder_with_default_dialect():
    data = msgpack.dumps([738785, 738786])
    decoder = MessagePackDecoder(List[date], default_dialect=MyDialect)
    assert decoder.decode(data) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_encoder_with_default_dialect():
    data = msgpack.dumps([738785, 738786])
    encoder = MessagePackEncoder(List[date], default_dialect=MyDialect)
    assert encoder.encode([date(2023, 9, 22), date(2023, 9, 23)]) == data


def test_pre_decoder_func():
    data = msgpack.dumps(["2023-09-22", "2023-09-23"])
    calls = 0

    def pre_decoder_func(value):
        nonlocal calls
        calls += 1
        return msgpack.loads(value)

    decoder = MessagePackDecoder(
        List[date],
        pre_decoder_func=pre_decoder_func,
    )
    assert decoder.decode(data) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]
    assert calls == 1


def test_post_encoder_func():
    data = msgpack.dumps(["2023-09-22", "2023-09-23"])
    calls = 0

    def post_encoder_func(value):
        nonlocal calls
        calls += 1
        return msgpack.dumps(value)

    encoder = MessagePackEncoder(
        List[date],
        post_encoder_func=post_encoder_func,
    )
    assert (
        encoder.encode(
            [
                date(2023, 9, 22),
                date(2023, 9, 23),
            ]
        )
        == data
    )
    assert calls == 1



from __future__ import annotations

from dataclasses import dataclass

from .test_foobar import Bar, Foo


@dataclass
class Baz(Foo[int]):
    pass


def test_class_with_base_in_another_module():
    assert Bar(x=1).to_json() == '{"x": 1}'
    assert Baz(x=1).to_json() == '{"x": 1}'



from dataclasses import dataclass
from datetime import date, datetime, time
from typing import List, Optional

import tomli_w

from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

from mashumaro.dialect import Dialect
from mashumaro.mixins.toml import DataClassTOMLMixin

serialization_strategy = {
    datetime: {
        "serialize": lambda dt: dt.strftime("%Y/%m/%d/%H/%M/%S"),
        "deserialize": lambda s: datetime.strptime(s, "%Y/%m/%d/%H/%M/%S"),
    },
    date: {
        "serialize": date.toordinal,
        "deserialize": date.fromordinal,
    },
    time: {
        "serialize": lambda t: t.strftime("%H/%M/%S"),
        "deserialize": lambda s: datetime.strptime(s, "%H/%M/%S").time(),
    },
}


class MyDialect(Dialect):
    serialization_strategy = serialization_strategy


@dataclass
class InnerDataClassWithOptionalField(DataClassTOMLMixin):
    x: Optional[int] = None


def test_data_class_toml_mixin():
    assert DataClassTOMLMixin.from_toml("") is None
    assert DataClassTOMLMixin().to_toml() is None


def test_to_toml():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        x: List[int]

    dumped = tomli_w.dumps({"x": [1, 2, 3]})
    assert DataClass([1, 2, 3]).to_toml() == dumped


def test_from_toml():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        x: List[int]

    dumped = tomli_w.dumps({"x": [1, 2, 3]})
    assert DataClass.from_toml(dumped) == DataClass([1, 2, 3])


def test_toml_with_serialization_strategy():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]

        class Config(BaseConfig):
            serialization_strategy = serialization_strategy

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    instance = DataClass([_datetime], [_date], [_time])
    dict_dumped = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
    }
    toml_dumped = tomli_w.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert DataClass.from_toml(toml_dumped) == instance
    assert instance.to_dict() == dict_dumped
    assert instance.to_toml() == toml_dumped


def test_toml_with_dialect():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]

        class Config(BaseConfig):
            dialect = MyDialect

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    instance = DataClass([_datetime], [_date], [_time])
    dict_dumped = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
    }
    toml_dumped = tomli_w.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert DataClass.from_toml(toml_dumped) == instance
    assert instance.to_dict() == dict_dumped
    assert instance.to_toml() == toml_dumped


def test_toml_with_dialect_support():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        datetime: List[datetime]
        date: List[date]
        time: List[time]

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]

    _datetime = datetime(2022, 10, 12, 12, 54, 30)
    _date = date(2022, 10, 12)
    _time = time(12, 54, 30)
    _datetime_dumped = _datetime.strftime("%Y/%m/%d/%H/%M/%S")
    _date_dumped = _date.toordinal()
    _time_dumped = _time.strftime("%H/%M/%S")
    instance = DataClass([_datetime], [_date], [_time])
    dict_dumped = {
        "datetime": [_datetime.isoformat()],
        "date": [_date.isoformat()],
        "time": [_time.isoformat()],
    }
    dict_dumped_dialect = {
        "datetime": [_datetime_dumped],
        "date": [_date_dumped],
        "time": [_time_dumped],
    }
    toml_dumped = tomli_w.dumps(
        {"datetime": [_datetime], "date": [_date], "time": [_time]}
    )
    toml_dumped_dialect = tomli_w.dumps(
        {
            "datetime": [_datetime_dumped],
            "date": [_date_dumped],
            "time": [_time_dumped],
        }
    )
    assert DataClass.from_dict(dict_dumped) == instance
    assert (
        DataClass.from_dict(dict_dumped_dialect, dialect=MyDialect) == instance
    )
    assert DataClass.from_toml(toml_dumped) == instance
    assert (
        DataClass.from_toml(toml_dumped_dialect, dialect=MyDialect) == instance
    )
    assert instance.to_dict() == dict_dumped
    assert instance.to_dict(dialect=MyDialect) == dict_dumped_dialect
    assert instance.to_toml() == toml_dumped
    assert instance.to_toml(dialect=MyDialect) == toml_dumped_dialect


def test_toml_omit_none():
    @dataclass
    class DataClass(DataClassTOMLMixin):
        x: Optional[InnerDataClassWithOptionalField] = None
        y: Optional[int] = None

    obj = DataClass()
    assert obj.to_dict() == {"x": None, "y": None}
    assert obj.to_toml() == ""
    obj = DataClass(InnerDataClassWithOptionalField())
    assert obj.to_dict() == {"x": {"x": None}, "y": None}
    assert obj.to_toml() == "[x]\n"
    assert DataClass.from_toml("[x]\n") == obj



from __future__ import annotations

from typing import ForwardRef, TypedDict

import pytest

from mashumaro.core.meta.helpers import get_function_arg_annotation
from mashumaro.jsonschema import build_json_schema
from mashumaro.jsonschema.models import (
    JSONObjectSchema,
    JSONSchema,
    JSONSchemaInstanceType,
)


class MyTypedDict(TypedDict):
    x: int


def test_jsonschema_generation_for_forward_refs():
    def foo(x: int, y: MyTypedDict): ...

    x_type = get_function_arg_annotation(foo, "x")
    assert isinstance(x_type, ForwardRef)
    assert build_json_schema(x_type).type is JSONSchemaInstanceType.INTEGER

    y_type = get_function_arg_annotation(foo, "y")
    assert isinstance(y_type, ForwardRef)
    assert build_json_schema(y_type) == JSONObjectSchema(
        type=JSONSchemaInstanceType.OBJECT,
        properties={"x": JSONSchema(type=JSONSchemaInstanceType.INTEGER)},
        additionalProperties=False,
        required=["x"],
    )



import collections
import collections.abc
import decimal
import fractions
import ipaddress
import os
import re
import uuid
from dataclasses import InitVar, dataclass, field
from datetime import date, datetime, time, timedelta, timezone
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from queue import Queue
from types import MappingProxyType
from typing import (
    Any,
    AnyStr,
    ChainMap,
    ClassVar,
    Counter,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Hashable,
    List,
    Mapping,
    MutableMapping,
    MutableSet,
    NewType,
    Optional,
    OrderedDict,
    Sequence,
    Set,
    Tuple,
)
from zoneinfo import ZoneInfo

import pytest
from typing_extensions import Final, LiteralString

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder, BasicEncoder
from mashumaro.codecs.basic import decode, encode
from mashumaro.config import BaseConfig
from mashumaro.exceptions import (
    InvalidFieldValue,
    MissingField,
    UnserializableDataError,
    UnserializableField,
)
from mashumaro.types import (
    GenericSerializableType,
    RoundedDecimal,
    SerializableType,
    SerializationStrategy,
)
from tests.entities import (
    MyUntypedNamedTupleWithDefaults,
    TDefaultInt,
    TypedDictWithReadOnly,
)

from .conftest import add_unpack_method
from .entities import (
    CustomPath,
    DataClassWithoutMixin,
    GenericNamedTuple,
    GenericSerializableList,
    GenericSerializableTypeDataClass,
    GenericTypedDict,
    MutableString,
    MyDataClass,
    MyDataClassWithUnion,
    MyDatetimeNewType,
    MyEnum,
    MyFlag,
    MyIntEnum,
    MyIntFlag,
    MyNamedTuple,
    MyNamedTupleWithDefaults,
    MyNamedTupleWithOptional,
    MyNativeStrEnum,
    MyStrEnum,
    MyUntypedNamedTuple,
    SerializableTypeDataClass,
    T,
    T_Optional_int,
    TAny,
    TInt,
    TIntStr,
    TMyDataClass,
    TypedDictOptionalKeys,
    TypedDictOptionalKeysWithOptional,
    TypedDictRequiredAndOptionalKeys,
    TypedDictRequiredKeys,
    TypedDictRequiredKeysWithOptional,
)
from .utils import same_types

NoneType = type(None)


class Fixture:
    T = 123
    T_INT = 123
    T_INT_STR = 123
    ANY = 123
    INT = 123
    FLOAT = 1.23
    BOOL = True
    LIST = [1, 2, 3]
    TUPLE = (1,)
    TUPLE_DUMPED = [1]
    DEQUE = collections.deque([1, 2, 3])
    SET = {1, 2, 3}
    FROZEN_SET = frozenset([1, 2, 3])
    CHAIN_MAP = collections.ChainMap({"a": 1, "b": 2}, {"c": 3, "d": 4})
    MAPS_LIST = [{"a": 1, "b": 2}, {"c": 3, "d": 4}]
    DICT = {"a": 1, "b": 2}
    ORDERED_DICT = collections.OrderedDict(a=1, b=2)
    DEFAULT_DICT = collections.defaultdict(int, a=1, b=2)
    MAPPING_PROXY = MappingProxyType(DICT)
    DEFAULT_NONE_DICT = collections.defaultdict(None, a=1, b=2)
    COUNTER: Counter[str] = collections.Counter(a=1, b=2)
    BYTES = b"123"
    BYTES_BASE64 = "MTIz\n"
    BYTE_ARRAY = bytearray(b"123")
    STR = "123"
    ENUM = MyEnum.a
    INT_ENUM = MyIntEnum.a
    STR_ENUM = MyStrEnum.a
    STR_ENUM_NATIVE = MyNativeStrEnum.a
    FLAG = MyFlag.a
    INT_FLAG = MyIntFlag.a
    DATA_CLASS = MyDataClass(a=1, b=2)
    T_DATA_CLASS = MyDataClass(a=1, b=2)
    DATA_CLASS_WITH_UNION = MyDataClassWithUnion(a=1, b=2)
    NONE = None
    DATETIME = datetime(2018, 10, 29, 12, 46, 55, 308495)
    DATETIME_STR = "2018-10-29T12:46:55.308495"
    DATE = DATETIME.date()
    DATE_STR = "2018-10-29"
    TIME = DATETIME.time()
    TIME_STR = "12:46:55.308495"
    TIMEDELTA = timedelta(3.14159265358979323846)
    TIMEZONE = timezone(timedelta(hours=3))
    UUID = uuid.UUID("3c25dd74-f208-46a2-9606-dd3919e975b7")
    UUID_STR = "3c25dd74-f208-46a2-9606-dd3919e975b7"
    IP4ADDRESS_STR = "127.0.0.1"
    IP4ADDRESS = ipaddress.IPv4Address(IP4ADDRESS_STR)
    IP6ADDRESS_STR = "::1"
    IP6ADDRESS = ipaddress.IPv6Address(IP6ADDRESS_STR)
    IP4NETWORK_STR = "127.0.0.0/8"
    IP4NETWORK = ipaddress.IPv4Network(IP4NETWORK_STR)
    IP6NETWORK_STR = "::/128"
    IP6NETWORK = ipaddress.IPv6Network(IP6NETWORK_STR)
    IP4INTERFACE_STR = "192.168.1.1/24"
    IP4INTERFACE = ipaddress.IPv4Interface(IP4INTERFACE_STR)
    IP6INTERFACE_STR = "::1/128"
    IP6INTERFACE = ipaddress.IPv6Interface(IP6INTERFACE_STR)
    DECIMAL = decimal.Decimal("1.33")
    DECIMAL_STR = "1.33"
    FRACTION = fractions.Fraction("1/3")
    FRACTION_STR = "1/3"
    MUTABLE_STRING = MutableString(STR)
    MUTABLE_STRING_STR = STR
    CUSTOM_PATH = CustomPath("/a/b/c")
    CUSTOM_PATH_STR = "/a/b/c"
    CUSTOM_SERIALIZE = "_FOOBAR_"
    GENERIC_SERIALIZABLE_LIST_INT = GenericSerializableList([1, 2, 3])
    GENERIC_SERIALIZABLE_LIST_STR = GenericSerializableList(["a", "b", "c"])
    LITERAL_STRING = "foo"
    PATTERN_STR = re.compile("[a-z]+")
    PATTERN_BYTES = re.compile(b"[a-z]+")


inner_values = [
    (T, Fixture.T, Fixture.T),
    (TInt, Fixture.T_INT, Fixture.T_INT),
    (TIntStr, Fixture.T_INT_STR, Fixture.T_INT_STR),
    (TDefaultInt, Fixture.T_INT, Fixture.T_INT),
    (Any, Fixture.ANY, Fixture.ANY),
    (TAny, Fixture.ANY, Fixture.ANY),
    (int, Fixture.INT, Fixture.INT),
    (float, Fixture.FLOAT, Fixture.FLOAT),
    (bool, Fixture.BOOL, Fixture.BOOL),
    (List[int], Fixture.LIST, Fixture.LIST),
    (list[int], Fixture.LIST, Fixture.LIST),
    (List, Fixture.LIST, Fixture.LIST),
    (list, Fixture.LIST, Fixture.LIST),
    (Deque[int], Fixture.DEQUE, Fixture.LIST),
    (collections.deque[int], Fixture.DEQUE, Fixture.LIST),
    (Deque, Fixture.DEQUE, Fixture.LIST),
    (collections.deque, Fixture.DEQUE, Fixture.LIST),
    (Tuple[int], Fixture.TUPLE, Fixture.TUPLE_DUMPED),
    (tuple[int], Fixture.TUPLE, Fixture.TUPLE_DUMPED),
    (Tuple, Fixture.TUPLE, Fixture.TUPLE_DUMPED),
    (tuple, Fixture.TUPLE, Fixture.TUPLE_DUMPED),
    (Set[int], Fixture.SET, Fixture.LIST),
    (set[int], Fixture.SET, Fixture.LIST),
    (Set, Fixture.SET, Fixture.LIST),
    (set, Fixture.SET, Fixture.LIST),
    (FrozenSet[int], Fixture.FROZEN_SET, Fixture.LIST),
    (frozenset[int], Fixture.FROZEN_SET, Fixture.LIST),
    (FrozenSet, Fixture.FROZEN_SET, Fixture.LIST),
    (frozenset, Fixture.FROZEN_SET, Fixture.LIST),
    (collections.abc.Set[int], Fixture.SET, Fixture.LIST),
    (collections.abc.Set, Fixture.SET, Fixture.LIST),
    (collections.abc.MutableSet[int], Fixture.SET, Fixture.LIST),
    (collections.abc.MutableSet, Fixture.SET, Fixture.LIST),
    (ChainMap[str, int], Fixture.CHAIN_MAP, Fixture.MAPS_LIST),
    (
        collections.ChainMap[str, int],
        Fixture.CHAIN_MAP,
        Fixture.MAPS_LIST,
    ),
    (ChainMap, Fixture.CHAIN_MAP, Fixture.MAPS_LIST),
    (collections.ChainMap, Fixture.CHAIN_MAP, Fixture.MAPS_LIST),
    (Dict[str, int], Fixture.DICT, Fixture.DICT),
    (dict[str, int], Fixture.DICT, Fixture.DICT),
    (Dict, Fixture.DICT, Fixture.DICT),
    (dict, Fixture.DICT, Fixture.DICT),
    (Mapping[str, int], Fixture.DICT, Fixture.DICT),
    (collections.abc.Mapping[str, int], Fixture.DICT, Fixture.DICT),
    (Mapping, Fixture.DICT, Fixture.DICT),
    (collections.abc.Mapping, Fixture.DICT, Fixture.DICT),
    (OrderedDict[str, int], Fixture.ORDERED_DICT, Fixture.DICT),
    (
        collections.OrderedDict[str, int],
        Fixture.ORDERED_DICT,
        Fixture.DICT,
    ),
    (OrderedDict, Fixture.ORDERED_DICT, Fixture.DICT),
    (collections.OrderedDict, Fixture.ORDERED_DICT, Fixture.DICT),
    (DefaultDict[str, int], Fixture.DEFAULT_DICT, Fixture.DICT),
    (
        collections.defaultdict[str, int],
        Fixture.DEFAULT_DICT,
        Fixture.DICT,
    ),
    (DefaultDict, Fixture.DEFAULT_NONE_DICT, Fixture.DICT),
    (collections.defaultdict, Fixture.DEFAULT_NONE_DICT, Fixture.DICT),
    (Counter[str], Fixture.COUNTER, Fixture.DICT),
    (collections.Counter[str], Fixture.COUNTER, Fixture.DICT),
    (Counter, Fixture.COUNTER, Fixture.DICT),
    (collections.Counter, Fixture.COUNTER, Fixture.DICT),
    (MutableMapping[str, int], Fixture.DICT, Fixture.DICT),
    (
        collections.abc.MutableMapping[str, int],
        Fixture.DICT,
        Fixture.DICT,
    ),
    (MutableMapping, Fixture.DICT, Fixture.DICT),
    (collections.abc.MutableMapping, Fixture.DICT, Fixture.DICT),
    (MappingProxyType[str, int], Fixture.MAPPING_PROXY, Fixture.DICT),
    (MappingProxyType, Fixture.MAPPING_PROXY, Fixture.DICT),
    (Sequence[int], Fixture.LIST, Fixture.LIST),
    (collections.abc.Sequence[int], Fixture.LIST, Fixture.LIST),
    (Sequence, Fixture.LIST, Fixture.LIST),
    (collections.abc.Sequence, Fixture.LIST, Fixture.LIST),
    (collections.abc.MutableSequence[int], Fixture.LIST, Fixture.LIST),
    (collections.abc.MutableSequence, Fixture.LIST, Fixture.LIST),
    (bytes, Fixture.BYTES, Fixture.BYTES_BASE64),
    (bytearray, Fixture.BYTE_ARRAY, Fixture.BYTES_BASE64),
    (str, Fixture.STR, Fixture.STR),
    (MyEnum, Fixture.ENUM, Fixture.ENUM.value),
    (MyStrEnum, Fixture.STR_ENUM, Fixture.STR_ENUM.value),
    (MyNativeStrEnum, Fixture.STR_ENUM_NATIVE, Fixture.STR_ENUM_NATIVE.value),
    (MyIntEnum, Fixture.INT_ENUM, Fixture.INT_ENUM.value),
    (MyFlag, Fixture.FLAG, Fixture.FLAG.value),
    (MyIntFlag, Fixture.INT_FLAG, Fixture.INT_FLAG.value),
    (MyDataClass, Fixture.DATA_CLASS, Fixture.DICT),
    (TMyDataClass, Fixture.T_DATA_CLASS, Fixture.DICT),
    (MyDataClassWithUnion, Fixture.DATA_CLASS_WITH_UNION, Fixture.DICT),
    (NoneType, Fixture.NONE, Fixture.NONE),
    (datetime, Fixture.DATETIME, Fixture.DATETIME_STR),
    (date, Fixture.DATE, Fixture.DATE_STR),
    (time, Fixture.TIME, Fixture.TIME_STR),
    (timedelta, Fixture.TIMEDELTA, Fixture.TIMEDELTA.total_seconds()),
    (timezone, Fixture.TIMEZONE, "UTC+03:00"),
    (ZoneInfo, ZoneInfo("Europe/Moscow"), "Europe/Moscow"),
    (uuid.UUID, Fixture.UUID, Fixture.UUID_STR),
    (ipaddress.IPv4Address, Fixture.IP4ADDRESS, Fixture.IP4ADDRESS_STR),
    (ipaddress.IPv6Address, Fixture.IP6ADDRESS, Fixture.IP6ADDRESS_STR),
    (ipaddress.IPv4Network, Fixture.IP4NETWORK, Fixture.IP4NETWORK_STR),
    (ipaddress.IPv6Network, Fixture.IP6NETWORK, Fixture.IP6NETWORK_STR),
    (ipaddress.IPv4Interface, Fixture.IP4INTERFACE, Fixture.IP4INTERFACE_STR),
    (ipaddress.IPv6Interface, Fixture.IP6INTERFACE, Fixture.IP6INTERFACE_STR),
    (decimal.Decimal, Fixture.DECIMAL, Fixture.DECIMAL_STR),
    (fractions.Fraction, Fixture.FRACTION, Fixture.FRACTION_STR),
    (MutableString, Fixture.MUTABLE_STRING, Fixture.MUTABLE_STRING_STR),
    (CustomPath, Fixture.CUSTOM_PATH, Fixture.CUSTOM_PATH_STR),
    (
        GenericSerializableList[int],
        Fixture.GENERIC_SERIALIZABLE_LIST_INT,
        [3, 4, 5],
    ),
    (
        GenericSerializableList[str],
        Fixture.GENERIC_SERIALIZABLE_LIST_STR,
        ["_a", "_b", "_c"],
    ),
    (MyDatetimeNewType, Fixture.DATETIME, Fixture.DATETIME_STR),
    (LiteralString, Fixture.LITERAL_STRING, Fixture.LITERAL_STRING),
    (re.Pattern, Fixture.PATTERN_STR, Fixture.PATTERN_STR.pattern),
    (
        re.Pattern[str],
        Fixture.PATTERN_STR,
        Fixture.PATTERN_STR.pattern,
    ),
    (
        re.Pattern[bytes],
        Fixture.PATTERN_BYTES,
        Fixture.PATTERN_BYTES.pattern,
    ),
]

if os.name == "posix":
    inner_values.extend(
        [
            (Path, Path("/a/b/c"), "/a/b/c"),
            (PurePath, PurePath("/a/b/c"), "/a/b/c"),
            (PosixPath, PosixPath("/a/b/c"), "/a/b/c"),
            (PurePosixPath, PurePosixPath("/a/b/c"), "/a/b/c"),
            (os.PathLike, PurePosixPath("/a/b/c"), "/a/b/c"),
        ]
    )
else:
    inner_values.extend(
        [
            (Path, Path("/a/b/c"), "\\a\\b\\c"),
            (PurePath, PurePath("/a/b/c"), "\\a\\b\\c"),
            (WindowsPath, WindowsPath("C:/Windows"), "C:\\Windows"),
            (
                PureWindowsPath,
                PureWindowsPath("C:/Program Files"),
                "C:\\Program Files",
            ),
            (os.PathLike, PureWindowsPath("/a/b/c"), "\\a\\b\\c"),
        ]
    )


hashable_inner_values = [
    (type_, value, value_dumped)
    for type_, value, value_dumped in inner_values
    if isinstance(value, Hashable) and isinstance(value_dumped, Hashable)
]


generic_sequence_types = [
    List,
    Deque,
    Tuple,
    Set,
    FrozenSet,
    list,
    collections.deque,
    tuple,
    set,
    frozenset,
    collections.abc.Set,
    collections.abc.MutableSet,
    collections.Counter,
    collections.abc.Sequence,
    collections.abc.MutableSequence,
]
generic_mapping_types = [
    Dict,
    Mapping,
    OrderedDict,
    DefaultDict,
    MutableMapping,
    collections.ChainMap,
    dict,
    collections.abc.Mapping,
    collections.OrderedDict,
    collections.defaultdict,
    collections.abc.MutableMapping,
]


unsupported_field_types = [Queue]
unsupported_typing_primitives = [AnyStr]


x_factory_mapping = {
    List: list,
    list: list,
    Deque: collections.deque,
    collections.deque: collections.deque,
    Tuple: tuple,
    tuple: tuple,
    Set: set,
    set: set,
    FrozenSet: frozenset,
    frozenset: frozenset,
    MutableSet: set,
    collections.abc.MutableSet: set,
    Dict: lambda items: {k: v for k, v in items},
    dict: lambda items: {k: v for k, v in items},
    Mapping: lambda items: {k: v for k, v in items},
    collections.abc.Mapping: lambda items: {k: v for k, v in items},
    MutableMapping: lambda items: {k: v for k, v in items},
    collections.abc.MutableMapping: lambda items: {k: v for k, v in items},
    OrderedDict: lambda items: {k: v for k, v in items},
    collections.OrderedDict: lambda items: {k: v for k, v in items},
    DefaultDict: lambda items: {k: v for k, v in items},
    collections.defaultdict: lambda items: {k: v for k, v in items},
    Counter: lambda items: collections.Counter({k: v for k, v in items}),
    collections.Counter: lambda items: collections.Counter(
        {k: v for k, v in items}
    ),
    ChainMap: lambda items: collections.ChainMap(*({k: v} for k, v in items)),
    collections.ChainMap: lambda items: collections.ChainMap(
        *({k: v} for k, v in items)
    ),
}


# noinspection PyCallingNonCallable
def check_collection_generic(type_, value_info, x_values_number=3):
    x_type, x_value, x_value_dumped = value_info

    @dataclass
    class DataClass(DataClassDictMixin):
        x: type_[x_type]

    x_factory = x_factory_mapping[type_]
    x = x_factory([x_value for _ in range(x_values_number)])
    instance = DataClass(x)
    dumped = {
        "x": list(x_factory([x_value_dumped for _ in range(x_values_number)]))
    }
    instance_dumped = instance.to_dict()
    instance_loaded = DataClass.from_dict(dumped)
    assert instance_dumped == dumped
    assert instance_loaded == instance
    assert same_types(instance_dumped, dumped)
    assert same_types(instance_loaded.x, x)


# noinspection PyCallingNonCallable
def check_mapping_generic(type_, key_info, value_info):
    k_type, k_value, k_value_dumped = key_info
    v_type, v_value, v_value_dumped = value_info

    if type_ is Counter:
        x_type = type_[k_type]
    else:
        x_type = type_[k_type, v_type]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: x_type

    x_factory = x_factory_mapping[type_]
    if type_ is Counter:
        x = x_factory([(k_value, 1) for _ in range(3)])
    else:
        x = x_factory([(k_value, v_value) for _ in range(3)])
    instance = DataClass(x)
    k_dumped = k_value_dumped
    v_dumped = v_value_dumped
    if type_ is ChainMap:
        dumped = {
            "x": x_factory([(k_dumped, v_dumped) for _ in range(3)]).maps
        }
    elif type_ is Counter:
        dumped = {"x": x_factory([(k_dumped, 1) for _ in range(3)])}
    else:
        dumped = {"x": x_factory([(k_dumped, v_dumped) for _ in range(3)])}
    instance_dumped = instance.to_dict()
    instance_loaded = DataClass.from_dict(dumped)
    assert instance_dumped == dumped
    assert instance_loaded == instance
    assert same_types(instance_dumped, dumped)
    assert same_types(instance_loaded.x, x)


@pytest.mark.parametrize("value_info", inner_values)
def test_one_level(value_info):
    x_type, x_value, x_value_dumped = value_info

    @dataclass
    class DataClass(DataClassDictMixin):
        x: x_type

    instance = DataClass(x_value)
    dumped = {"x": x_value_dumped}
    instance_dumped = instance.to_dict()
    instance_loaded = DataClass.from_dict(dumped)
    assert instance_dumped == dumped
    assert instance_loaded == instance
    assert same_types(instance_dumped, dumped)
    assert same_types(instance_loaded.x, x_value)


@pytest.mark.parametrize("value_info", inner_values)
def test_with_generic_list(value_info):
    check_collection_generic(List, value_info)


@pytest.mark.parametrize("value_info", inner_values)
def test_with_generic_deque(value_info):
    check_collection_generic(Deque, value_info)


@pytest.mark.parametrize("value_info", inner_values)
def test_with_generic_tuple(value_info):
    check_collection_generic(Tuple, value_info, 1)


@pytest.mark.parametrize("value_info", hashable_inner_values)
def test_with_generic_set(value_info):
    check_collection_generic(Set, value_info)


@pytest.mark.parametrize("value_info", hashable_inner_values)
def test_with_generic_frozenset(value_info):
    check_collection_generic(FrozenSet, value_info)


@pytest.mark.parametrize("value_info", hashable_inner_values)
def test_with_generic_mutable_set(value_info):
    check_collection_generic(MutableSet, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_dict(key_info, value_info):
    check_mapping_generic(Dict, key_info, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_mapping(key_info, value_info):
    check_mapping_generic(Mapping, key_info, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_ordered_dict(key_info, value_info):
    check_mapping_generic(OrderedDict, key_info, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_counter(key_info, value_info):
    check_mapping_generic(Counter, key_info, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_mutable_mapping(key_info, value_info):
    check_mapping_generic(MutableMapping, key_info, value_info)


@pytest.mark.parametrize("value_info", inner_values)
@pytest.mark.parametrize("key_info", hashable_inner_values)
def test_with_generic_chain_map(key_info, value_info):
    check_mapping_generic(ChainMap, key_info, value_info)


@pytest.mark.parametrize("x_type", unsupported_field_types)
@pytest.mark.parametrize("generic_type", generic_sequence_types)
def test_unsupported_generic_field_types(x_type, generic_type):
    with pytest.raises(UnserializableField):

        @dataclass
        class _(DataClassDictMixin):
            # noinspection PyTypeChecker
            x: generic_type[x_type]

    with add_unpack_method:
        with pytest.raises(UnserializableField):

            @dataclass
            class _(DataClassDictMixin):
                # noinspection PyTypeChecker
                x: generic_type[x_type]


@pytest.mark.parametrize("x_type", unsupported_typing_primitives)
@pytest.mark.parametrize("generic_type", generic_sequence_types)
def test_unsupported_generic_typing_primitives(x_type, generic_type):
    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            # noinspection PyTypeChecker
            x: generic_type[x_type]

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                # noinspection PyTypeChecker
                x: generic_type[x_type]


@pytest.mark.parametrize("x_type", unsupported_field_types)
def test_unsupported_field_types(x_type):
    with pytest.raises(UnserializableField):

        @dataclass
        class _(DataClassDictMixin):
            x: x_type

    with add_unpack_method:
        with pytest.raises(UnserializableField):

            @dataclass
            class _(DataClassDictMixin):
                x: x_type


@pytest.mark.parametrize("x_type", unsupported_typing_primitives)
def test_unsupported_typing_primitives(x_type):
    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            x: x_type

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                x: x_type


@pytest.mark.parametrize("generic_type", generic_mapping_types)
def test_data_class_as_mapping_key(generic_type):
    @dataclass
    class Key(DataClassDictMixin):
        pass

    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            x: generic_type[Key, int]

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                x: generic_type[Key, int]


def test_data_class_as_mapping_key_for_counter():
    @dataclass
    class Key(DataClassDictMixin):
        pass

    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            x: Counter[Key]

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                x: Counter[Key]


def test_data_class_as_chain_map_key():
    @dataclass
    class Key(DataClassDictMixin):
        pass

    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            x: ChainMap[Key, int]

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                x: ChainMap[Key, int]


@pytest.mark.parametrize("value_info", inner_values)
def test_with_any(value_info):
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Any

    x = value_info[1]
    dumped = {"x": x}
    instance = DataClass(x)
    instance_dumped = instance.to_dict()
    instance_loaded = DataClass.from_dict(dumped)
    assert instance_dumped == dumped
    assert instance_loaded == instance
    assert same_types(instance_dumped, dumped)
    assert same_types(instance_loaded.x, x)


@pytest.mark.parametrize("value_info", inner_values)
def test_with_optional(value_info):
    x_type, x_value, x_value_dumped = value_info

    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[x_type] = None

    for instance in [DataClass(x_value), DataClass()]:
        if instance.x is None:
            v_dumped = None
        else:
            v_dumped = x_value_dumped
        dumped = {"x": v_dumped}
        instance_dumped = instance.to_dict()
        instance_loaded = DataClass.from_dict(dumped)
        assert instance_dumped == dumped
        assert instance_loaded == instance
        assert same_types(instance_dumped, dumped)
        assert same_types(instance_loaded.x, instance.x)


def test_raises_missing_field():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int

    with pytest.raises(MissingField):
        DataClass.from_dict({})


def test_empty_dataclass():
    @dataclass
    class DataClass(DataClassDictMixin):
        pass

    assert DataClass().to_dict() == {}
    assert type(DataClass.from_dict({})) is DataClass
    assert DataClass.from_dict({}).__dict__ == {}


def test_weird_field_type():
    with pytest.raises(UnserializableDataError):

        @dataclass
        class _(DataClassDictMixin):
            x: 123

    with add_unpack_method:
        with pytest.raises(UnserializableDataError):

            @dataclass
            class _(DataClassDictMixin):
                x: 123


@pytest.mark.parametrize(
    "rounding", [None, decimal.ROUND_UP, decimal.ROUND_DOWN]
)
@pytest.mark.parametrize("places", [None, 1, 2])
def test_rounded_decimal(places, rounding):
    @dataclass
    class DataClass(DataClassDictMixin):
        x: decimal.Decimal

        class Config(BaseConfig):
            serialization_strategy = {
                decimal.Decimal: RoundedDecimal(places, rounding)
            }

    digit = decimal.Decimal(0.35)
    if places is not None:
        exp = decimal.Decimal((0, (1,), -places))
        quantized = digit.quantize(exp, rounding)
    else:
        quantized = digit
    assert DataClass(digit).to_dict() == {"x": str(quantized)}
    assert DataClass.from_dict({"x": str(quantized)}) == DataClass(x=quantized)


def test_serializable_type():
    with pytest.raises(NotImplementedError):
        # noinspection PyTypeChecker
        SerializableType._serialize(None)
    with pytest.raises(NotImplementedError):
        SerializableType._deserialize(None)


def test_serialization_strategy():
    with pytest.raises(NotImplementedError):
        # noinspection PyTypeChecker
        SerializationStrategy.serialize(None, None)
    with pytest.raises(NotImplementedError):
        # noinspection PyTypeChecker
        SerializationStrategy.deserialize(None, None)


def test_class_vars():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: ClassVar[int] = None

    assert DataClass().to_dict() == {}
    assert DataClass.from_dict({}) == DataClass()


def test_final():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Final[int] = 42

    assert DataClass().to_dict() == {"x": 42}
    assert DataClass(42).to_dict() == {"x": 42}
    assert DataClass(33).to_dict() == {"x": 33}
    assert DataClass.from_dict({}) == DataClass()
    assert DataClass.from_dict({"x": 42}) == DataClass()
    assert DataClass.from_dict({"x": 33}) == DataClass(33)


def test_init_vars():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: InitVar[int] = None
        y: int = None

        def __post_init__(self, x: int):
            if self.y is None and x is not None:
                self.y = x

    assert DataClass().to_dict() == {"y": None}
    assert DataClass(x=1).to_dict() == {"y": 1}
    assert DataClass.from_dict({}) == DataClass()
    assert DataClass.from_dict({"x": 1}) == DataClass()


def test_dataclass_with_defaults():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int
        y: int = 1

    assert DataClass.from_dict({"x": 0}) == DataClass(x=0, y=1)


def test_dataclass_with_field_and_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(default=1)

    assert DataClass.from_dict({}) == DataClass(x=1)
    assert DataClass().to_dict() == {"x": 1}


def test_dataclass_with_field_and_no_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(metadata={})

    with pytest.raises(MissingField):
        assert DataClass.from_dict({})


def test_dataclass_with_field_and_default_factory():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: List[str] = field(default_factory=list)

    assert DataClass.from_dict({}) == DataClass(x=[])
    assert DataClass().to_dict() == {"x": []}


def test_derived_dataclass_with_ancestors_defaults():
    @dataclass
    class A:
        x: int
        y: int = 1

    @dataclass
    class B(A, DataClassDictMixin):
        z: int = 3

    @dataclass
    class C(B, DataClassDictMixin):
        y: int = 4

    @dataclass
    class D(C):
        pass

    assert B.from_dict({"x": 0}) == B(x=0, y=1, z=3)
    assert C.from_dict({"x": 0}) == C(x=0, y=4, z=3)
    assert D.from_dict({"x": 0}) == D(x=0, y=4, z=3)


def test_derived_dataclass_with_ancestors_and_field_with_default():
    @dataclass
    class A:
        x: int = field(default=1)

    @dataclass
    class B(A, DataClassDictMixin):
        y: int = field(default=2)

    @dataclass
    class C(B, DataClassDictMixin):
        z: int = field(default=3)

    @dataclass
    class D(C):
        pass

    assert B.from_dict({}) == B(x=1, y=2)
    assert C.from_dict({}) == C(x=1, y=2, z=3)
    assert D.from_dict({}) == D(x=1, y=2, z=3)


def test_derived_dataclass_with_ancestors_and_field_with_default_factory():
    @dataclass
    class A:
        x: List[int] = field(default_factory=lambda: [1])

    @dataclass
    class B(A, DataClassDictMixin):
        y: List[int] = field(default_factory=lambda: [2])

    @dataclass
    class C(B, DataClassDictMixin):
        z: List[int] = field(default_factory=lambda: [3])

    @dataclass
    class D(C):
        pass

    assert B.from_dict({}) == B(x=[1], y=[2])
    assert C.from_dict({}) == C(x=[1], y=[2], z=[3])
    assert D.from_dict({}) == D(x=[1], y=[2], z=[3])


def test_invalid_field_value_deserialization():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "bad_value"})


def test_invalid_field_value_deserialization_with_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = 1

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "bad_value"})


def test_invalid_field_value_deserialization_with_rounded_decimal():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: decimal.Decimal

        class Config(BaseConfig):
            serialization_strategy = {decimal.Decimal: RoundedDecimal()}

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "bad_value"})


def test_invalid_field_value_deserialization_with_rounded_decimal_with_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: decimal.Decimal = Fixture.DECIMAL

        class Config(BaseConfig):
            serialization_strategy = {decimal.Decimal: RoundedDecimal()}

    with pytest.raises(InvalidFieldValue):
        DataClass.from_dict({"x": "bad_value"})


@pytest.mark.parametrize(
    "value_info",
    [
        v
        for v in inner_values
        if v[0] not in [MyDataClass, NoneType, MutableString]
    ],
)
def test_serialize_deserialize_options(value_info):
    x_type, x_value, x_value_dumped = value_info

    @dataclass
    class DataClass(DataClassDictMixin):
        x: x_type = field(
            metadata={
                "serialize": lambda v: Fixture.CUSTOM_SERIALIZE,
                "deserialize": lambda v: (
                    x_value
                    if v == Fixture.CUSTOM_SERIALIZE
                    else f"!{Fixture.CUSTOM_SERIALIZE}"
                ),
            }
        )

    instance = DataClass(x_value)
    v_dumped = Fixture.CUSTOM_SERIALIZE
    dumped = {"x": v_dumped}
    instance_dumped = instance.to_dict()
    instance_loaded = DataClass.from_dict(dumped)
    assert instance_dumped == dumped
    assert instance_loaded == instance
    assert same_types(instance_dumped, dumped)
    assert same_types(instance_loaded.x, x_value)


def test_dataclass_field_without_mixin():
    @dataclass
    class DataClass(DataClassDictMixin):
        p: DataClassWithoutMixin

    obj = DataClass(DataClassWithoutMixin(42))
    assert DataClass.from_dict({"p": {"i": "42"}}) == obj
    assert obj.to_dict() == {"p": {"i": 42}}


def test_serializable_type_dataclass():
    @dataclass
    class DataClass(DataClassDictMixin):
        s: SerializableTypeDataClass

    s_value = SerializableTypeDataClass(a=9, b=9)
    assert DataClass.from_dict({"s": {"a": 10, "b": 10}}) == DataClass(s_value)
    assert DataClass(s_value).to_dict() == {"s": {"a": 10, "b": 10}}


def test_optional_inside_collection():
    @dataclass
    class DataClass(DataClassDictMixin):
        l: List[Optional[int]]
        d: Dict[Optional[int], Optional[int]]

    d = {"l": [1, None, 2], "d": {1: 1, 2: None, None: 3}}
    o = DataClass.from_dict(d)
    assert o == DataClass(l=[1, None, 2], d={1: 1, 2: None, None: 3})
    assert o.to_dict() == d


def test_bound_type_var_inside_collection():
    @dataclass
    class DataClass(DataClassDictMixin):
        l: List[T_Optional_int]
        d: Dict[T_Optional_int, T_Optional_int]

    d = {"l": [1, None, 2], "d": {1: 1, 2: None, None: 3}}
    o = DataClass.from_dict(d)
    assert o == DataClass(l=[1, None, 2], d={1: 1, 2: None, None: 3})
    assert o.to_dict() == d


def test_generic_serializable_type():
    with pytest.raises(NotImplementedError):
        # noinspection PyTypeChecker
        GenericSerializableType._serialize(None, None)
    with pytest.raises(NotImplementedError):
        GenericSerializableType._deserialize(None, None)


def test_generic_serializable_type_dataclass():
    @dataclass
    class DataClass(DataClassDictMixin):
        s: GenericSerializableTypeDataClass

    s_value = GenericSerializableTypeDataClass(a=9, b=9)
    assert DataClass.from_dict({"s": {"a": 10, "b": 10}}) == DataClass(s_value)
    assert DataClass(s_value).to_dict() == {"s": {"a": 10, "b": 10}}


def test_dataclass_with_different_tuples():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple
        c: Tuple[int]
        d: Tuple[int, float, int]
        e: Tuple[int, ...]

    obj = DataClass(a=(1, "2", 3.0), c=(1,), d=(1, 2.0, 3), e=(1, 2, 3))
    assert (
        DataClass.from_dict(
            {
                "a": [1, "2", 3.0],
                "c": [1, 2, 3],
                "d": ["1", "2.0", "3"],
                "e": [1, 2, 3],
            }
        )
        == obj
    )
    assert obj.to_dict() == {
        "a": [1, "2", 3.0],
        "c": [1],
        "d": [1, 2.0, 3],
        "e": [1, 2, 3],
    }


def test_dataclass_with_empty_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Tuple[()]

    obj = DataClass(x=())
    assert (
        DataClass.from_dict(
            {
                "x": [1, 2, 3],
            }
        )
        == obj
    )
    assert obj.to_dict() == {
        "x": [],
    }


def test_dataclass_with_typed_dict_required_keys():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictRequiredKeys

    for data in ({}, {"int": 1}, {"float": 1.0}):
        with pytest.raises(InvalidFieldValue):
            DataClass.from_dict({"x": data})

    obj = DataClass(x={"int": 1, "float": 2.0})
    assert (
        DataClass.from_dict({"x": {"int": "1", "float": "2.0", "str": "str"}})
        == obj
    )
    assert obj.to_dict() == {"x": {"int": 1, "float": 2.0}}


def test_dataclass_with_typed_dict_optional_keys():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictOptionalKeys

    for data in ({}, {"int": 1}, {"float": 1.0}):
        assert DataClass.from_dict({"x": data}) == DataClass(x=data)

    obj = DataClass(x={"int": 1, "float": 2.0})
    assert (
        DataClass.from_dict({"x": {"int": "1", "float": "2.0", "str": "str"}})
        == obj
    )
    assert obj.to_dict() == {"x": {"int": 1, "float": 2.0}}


def test_dataclass_with_typed_dict_required_and_optional_keys():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictRequiredAndOptionalKeys

    for data in (
        {},
        {"str": "str"},
        {"int": 1},
        {"float": 1.0},
        {"int": 1, "str": "str"},
        {"float": 1.0, "str": "str"},
    ):
        with pytest.raises(InvalidFieldValue):
            DataClass.from_dict({"x": data})

    assert DataClass.from_dict(
        {"x": {"int": "1", "float": "2.0", "unknown": "unknown"}}
    ) == DataClass(x={"int": 1, "float": 2.0})
    assert DataClass.from_dict(
        {"x": {"int": "1", "float": "2.0", "str": "str"}}
    ) == DataClass(x={"int": 1, "float": 2.0, "str": "str"})
    assert DataClass(x={"int": 1, "float": 2.0}).to_dict() == {
        "x": {"int": 1, "float": 2.0}
    }
    assert DataClass(x={"int": 1, "float": 2.0, "str": "str"}).to_dict() == {
        "x": {"int": 1, "float": 2.0, "str": "str"}
    }


def test_dataclass_with_typed_dict_with_read_only_key():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictWithReadOnly

    assert DataClass.from_dict({"x": {"x": "42"}}) == DataClass({"x": 42})
    assert DataClass({"x": 42}).to_dict() == {"x": {"x": 42}}


def test_dataclass_with_named_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyNamedTuple

    obj = DataClass(x=MyNamedTuple(1, 2.0))
    assert DataClass.from_dict({"x": ["1", "2.0"]}) == obj
    assert obj.to_dict() == {"x": [1, 2.0]}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": ["1", "2.0"]}) == obj
    assert encoder.encode(obj) == {"x": [1, 2.0]}


def test_dataclass_with_named_tuple_with_defaults():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyNamedTupleWithDefaults

    obj = DataClass(x=MyNamedTupleWithDefaults())
    assert DataClass.from_dict({"x": ["1"]}) == obj
    assert obj.to_dict() == {"x": [1, 2.0]}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": ["1"]}) == obj
    assert encoder.encode(obj) == {"x": [1, 2.0]}


def test_dataclass_with_untyped_named_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyUntypedNamedTuple

    obj = DataClass(x=MyUntypedNamedTuple("1", "2.0"))
    assert DataClass.from_dict({"x": ["1", "2.0"]}) == obj
    assert obj.to_dict() == {"x": ["1", "2.0"]}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": ["1", "2.0"]}) == obj
    assert encoder.encode(obj) == {"x": ["1", "2.0"]}


def test_dataclass_with_untyped_named_tuple_with_defaults():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyUntypedNamedTupleWithDefaults

    obj = DataClass(x=MyUntypedNamedTupleWithDefaults(i="1"))
    assert DataClass.from_dict({"x": ["1"]}) == obj
    assert obj.to_dict() == {"x": ["1", 2.0]}


def test_data_class_with_none():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: None
        y: NoneType
        z: List[None]

    obj = DataClass(x=None, y=None, z=[None])
    assert DataClass.from_dict({"x": None, "y": None, "z": [None]}) == obj
    assert DataClass.from_dict({"x": 42, "y": "foo", "z": ["bar"]}) == obj
    assert obj.to_dict() == {"x": None, "y": None, "z": [None]}


def test_data_class_with_new_type_overridden():
    MyStr = NewType("MyStr", str)

    @dataclass
    class DataClass(DataClassDictMixin):
        x: str
        y: MyStr

        class Config(BaseConfig):
            serialization_strategy = {
                str: {
                    "serialize": lambda x: f"str_{x}",
                    "deserialize": lambda x: x[4:],
                },
                MyStr: {
                    "serialize": lambda x: f"MyStr_{x}",
                    "deserialize": lambda x: x[6:],
                },
            }

    instance = DataClass("a", MyStr("b"))
    assert DataClass.from_dict({"x": "str_a", "y": "MyStr_b"}) == instance
    assert instance.to_dict() == {"x": "str_a", "y": "MyStr_b"}


def test_tuple_with_optional():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Tuple[Optional[int], int] = field(default_factory=lambda: (None, 7))

    assert DataClass.from_dict({"x": [None, 42]}) == DataClass((None, 42))
    assert DataClass((None, 42)).to_dict() == {"x": [None, 42]}
    assert DataClass.from_dict({}) == DataClass((None, 7))
    assert DataClass().to_dict() == {"x": [None, 7]}


def test_tuple_with_optional_and_ellipsis():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Tuple[Optional[int], ...] = field(default_factory=lambda: (None, 7))

    assert DataClass.from_dict({"x": [None, 42]}) == DataClass((None, 42))
    assert DataClass((None, 42)).to_dict() == {"x": [None, 42]}
    assert DataClass.from_dict({}) == DataClass((None, 7))
    assert DataClass().to_dict() == {"x": [None, 7]}


def test_named_tuple_with_optional():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyNamedTupleWithOptional = field(
            default_factory=lambda: MyNamedTupleWithOptional(None, 7)
        )

    assert DataClass.from_dict({"x": [None, 42]}) == DataClass(
        MyNamedTupleWithOptional(None, 42)
    )
    assert DataClass(MyNamedTupleWithOptional(None, 42)).to_dict() == {
        "x": [None, 42]
    }
    assert DataClass.from_dict({}) == DataClass(
        MyNamedTupleWithOptional(None, 7)
    )
    assert DataClass().to_dict() == {"x": [None, 7]}


def test_unbound_generic_named_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericNamedTuple

    obj = DataClass(GenericNamedTuple("2023-01-22", 42))
    assert DataClass.from_dict({"x": ["2023-01-22", "42"]}) == obj
    assert obj.to_dict() == {"x": ["2023-01-22", 42]}


def test_bound_generic_named_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericNamedTuple[date]

    obj = DataClass(GenericNamedTuple(date(2023, 1, 22), 42))
    assert DataClass.from_dict({"x": ["2023-01-22", "42"]}) == obj
    assert obj.to_dict() == {"x": ["2023-01-22", 42]}


def test_typed_dict_required_keys_with_optional():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictRequiredKeysWithOptional

    obj = DataClass({"x": None, "y": 42})
    assert DataClass.from_dict({"x": {"x": None, "y": 42}}) == obj
    assert obj.to_dict() == {"x": {"x": None, "y": 42}}

    obj = DataClass({"x": 33, "y": 42})
    assert DataClass.from_dict({"x": {"x": 33, "y": 42}}) == obj
    assert obj.to_dict() == {"x": {"x": 33, "y": 42}}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": {"x": 33, "y": 42}}) == obj
    assert encoder.encode(obj) == {"x": {"x": 33, "y": 42}}


def test_typed_dict_optional_keys_with_optional():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: TypedDictOptionalKeysWithOptional

    obj = DataClass({"x": None, "y": 42})
    assert DataClass.from_dict({"x": {"x": None, "y": 42}}) == obj
    assert obj.to_dict() == {"x": {"x": None, "y": 42}}

    obj = DataClass({"x": 33, "y": 42})
    assert DataClass.from_dict({"x": {"x": 33, "y": 42}}) == obj
    assert obj.to_dict() == {"x": {"x": 33, "y": 42}}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": {"x": 33, "y": 42}}) == obj
    assert encoder.encode(obj) == {"x": {"x": 33, "y": 42}}


def test_unbound_generic_typed_dict():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericTypedDict

    obj = DataClass({"x": "2023-01-22", "y": 42})
    assert DataClass.from_dict({"x": {"x": "2023-01-22", "y": "42"}}) == obj
    assert obj.to_dict() == {"x": {"x": "2023-01-22", "y": 42}}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": {"x": "2023-01-22", "y": "42"}}) == obj
    assert encoder.encode(obj) == {"x": {"x": "2023-01-22", "y": 42}}


def test_bound_generic_typed_dict():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericTypedDict[date]

    obj = DataClass({"x": date(2023, 1, 22), "y": 42})
    assert DataClass.from_dict({"x": {"x": "2023-01-22", "y": "42"}}) == obj
    assert obj.to_dict() == {"x": {"x": "2023-01-22", "y": 42}}

    decoder = BasicDecoder(DataClass)
    encoder = BasicEncoder(DataClass)
    assert decoder.decode({"x": {"x": "2023-01-22", "y": "42"}}) == obj
    assert encoder.encode(obj) == {"x": {"x": "2023-01-22", "y": 42}}


def test_dataclass_with_init_false_field():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(init=False)

        def __post_init__(self):
            self.x = 42

    obj = DataClass()
    assert obj.to_dict() == {"x": 42}
    assert DataClass.from_dict({"x": 42}) == obj
    assert DataClass.from_dict({}) == obj


def test_dataclass_with_non_optional_none_value():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int
        y: int = None
        z: int = 42

    with pytest.raises(InvalidFieldValue) as e:
        DataClass.from_dict({"x": None})
    assert e.value.field_name == "x"

    obj = DataClass(x=42)
    assert DataClass.from_dict({"x": 42}) == obj
    assert obj.to_dict() == {"x": 42, "y": None, "z": 42}


def test_dataclass_with_optional_list_with_optional_ints():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[List[Optional[int]]]

    obj = DataClass(x=[42, None])
    assert DataClass.from_dict({"x": [42, None]}) == obj
    assert obj.to_dict() == {"x": [42, None]}


def test_dataclass_with_default_nan_and_inf_with_omit_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: float = float("nan")
        b: float = float("inf")
        c: float = float("-inf")

        class Config(BaseConfig):
            omit_default = True

    assert DataClass().to_dict() == {}
    assert DataClass(float("nan"), float("inf"), float("-inf")).to_dict() == {}
    assert (
        DataClass(float("nan"), float("+inf"), float("-inf")).to_dict() == {}
    )
    assert DataClass(float("inf"), float("-inf"), float("+inf")).to_dict() == {
        "a": float("inf"),
        "b": float("-inf"),
        "c": float("+inf"),
    }


def test_dataclass_with_default_int_flag_omit_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: MyIntFlag = MyIntFlag.a
        b: MyIntFlag = MyIntFlag.b

        class Config(BaseConfig):
            omit_default = True

    assert DataClass().to_dict() == {}
    assert DataClass(MyIntFlag.a, MyIntFlag.b).to_dict() == {}


@pytest.mark.parametrize("value_info", inner_values)
def test_decoder(value_info):
    x_type, x_value, x_value_dumped = value_info
    decoder = BasicDecoder(x_type)
    assert decoder.decode(x_value_dumped) == x_value
    assert decode(x_value_dumped, x_type) == x_value


@pytest.mark.parametrize("value_info", inner_values)
def test_encoder(value_info):
    x_type, x_value, x_value_dumped = value_info
    encoder = BasicEncoder(x_type)
    assert encoder.encode(x_value) == x_value_dumped
    assert encode(x_value, x_type) == x_value_dumped



from dataclasses import dataclass
from datetime import date, datetime

from typing_extensions import Annotated

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig


def test_annotated():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Annotated[date, None]

    obj = DataClass(date(2022, 2, 6))
    assert DataClass.from_dict({"x": "2022-02-06"}) == obj
    assert obj.to_dict() == {"x": "2022-02-06"}


def test_annotated_with_overridden_methods():
    @dataclass
    class DataClass(DataClassDictMixin):
        foo: Annotated[date, "foo"]
        bar: Annotated[date, "bar"]
        baz: Annotated[date, "baz"]

        class Config(BaseConfig):
            serialization_strategy = {
                Annotated[date, "foo"]: {
                    "serialize": date.toordinal,
                    "deserialize": date.fromordinal,
                },
                Annotated[date, "bar"]: {
                    "serialize": date.isoformat,
                    "deserialize": date.fromisoformat,
                },
                date: {
                    "serialize": lambda x: x.strftime("%Y%m%d"),
                    "deserialize": (
                        lambda x: datetime.strptime(x, "%Y%m%d").date()
                    ),
                },
            }

    obj = DataClass(
        foo=date(2023, 6, 12),
        bar=date(2023, 6, 12),
        baz=date(2023, 6, 12),
    )
    obj.foo.strftime("%Y%M%D")
    assert (
        DataClass.from_dict(
            {"foo": 738683, "bar": "2023-06-12", "baz": "20230612"}
        )
        == obj
    )
    assert obj.to_dict() == {
        "foo": 738683,
        "bar": "2023-06-12",
        "baz": "20230612",
    }



from dataclasses import dataclass, field
from datetime import date
from typing import Generic, List, Sequence, TypeVar

from mashumaro import DataClassDictMixin, field_options
from mashumaro.types import SerializationStrategy

T = TypeVar("T")


class TruncatedGenericListSerializationStrategy(
    SerializationStrategy, Generic[T]
):
    def serialize(self, value: List[T]) -> Sequence[T]:
        return value[:-1]

    def deserialize(self, value: Sequence[T]) -> Sequence[T]:
        return value[:-1]


class TruncatedAnnotatedDateListSerializationStrategy(
    SerializationStrategy, use_annotations=True
):
    def serialize(self, value) -> Sequence[date]:
        return value[:-1]

    def deserialize(self, value: Sequence[date]):
        return value[:-1]


class TruncatedDateListSerializationStrategy(SerializationStrategy):
    def serialize(self, value) -> Sequence[date]:
        return value[:-1]

    def deserialize(self, value: Sequence[date]):
        return value[:-1]


class TruncatedUnannotatedListSerializationStrategy(
    SerializationStrategy, use_annotations=True
):
    def serialize(self, value):
        return value[:-1]

    def deserialize(self, value):
        return value[:-1]


def test_generic_list_serialization_strategy():
    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: List[int]
        y: List[List[int]]

        class Config:
            serialization_strategy = {
                list: TruncatedGenericListSerializationStrategy()
            }

    obj = MyDataClass([1, 2, 3, 4, 5], [[1, 2], [3, 4]])
    assert obj.to_dict() == {"x": [1, 2, 3, 4], "y": [[1]]}
    assert MyDataClass.from_dict(
        {"x": ["1", "2", "3", "4"], "y": [[1, 2], [3, 4]]}
    ) == MyDataClass([1, 2, 3], [[1]])


def test_date_list_serialization_strategy_with_use_annotations():
    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: List[date]

        class Config:
            serialization_strategy = {
                list: TruncatedAnnotatedDateListSerializationStrategy()
            }

    obj = MyDataClass(
        [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    )
    assert obj.to_dict() == {
        "x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]
    }
    assert MyDataClass.from_dict(
        {"x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]}
    ) == MyDataClass([date(2023, 2, 12), date(2023, 2, 12), date(2023, 2, 12)])


def test_date_list_serialization_strategy_without_use_annotations():
    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: List[date]

        class Config:
            serialization_strategy = {
                list: TruncatedDateListSerializationStrategy()
            }

    obj = MyDataClass(
        [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    )
    assert obj.to_dict() == {
        "x": [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    }
    assert MyDataClass.from_dict(
        {"x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]}
    ) == MyDataClass(["2023-02-12", "2023-02-12", "2023-02-12"])


def test_date_list_serialization_strategy_use_annotations_without_annotations():
    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: List[date]

        class Config:
            serialization_strategy = {
                list: TruncatedUnannotatedListSerializationStrategy()
            }

    obj = MyDataClass(
        [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    )
    assert obj.to_dict() == {
        "x": [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    }
    assert MyDataClass.from_dict(
        {"x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]}
    ) == MyDataClass(["2023-02-12", "2023-02-12", "2023-02-12"])


def test_date_list_field_serialization_strategy_with_use_annotations():
    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: List[date] = field(
            metadata=field_options(
                serialization_strategy=(
                    TruncatedAnnotatedDateListSerializationStrategy()
                )
            )
        )

    obj = MyDataClass(
        [
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
            date(2023, 2, 12),
        ]
    )
    assert obj.to_dict() == {
        "x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]
    }
    assert MyDataClass.from_dict(
        {"x": ["2023-02-12", "2023-02-12", "2023-02-12", "2023-02-12"]}
    ) == MyDataClass([date(2023, 2, 12), date(2023, 2, 12), date(2023, 2, 12)])



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Optional, TypeVar

from mashumaro.mixins.json import DataClassJSONMixin

T = TypeVar("T")


@dataclass
class Foo(Generic[T], DataClassJSONMixin):
    x: T
    y: Optional[Foo[Any]]


@dataclass
class Bar(Foo): ...


def test_self_referenced_generic_no_max_recursion_error():
    obj = Bar(42, Foo(33, None))
    assert obj.to_dict() == {"x": 42, "y": {"x": 33, "y": None}}
    assert Bar.from_dict({"x": 42, "y": {"x": 33, "y": None}}) == obj
    assert obj.to_json() == '{"x": 42, "y": {"x": 33, "y": null}}'
    assert Bar.from_json('{"x": 42, "y": {"x": 33, "y": null}}') == obj
